# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
from datetime import datetime, timedelta
import numpy as np
import numpy.testing as nptest
import os
import sys
import pytest

import ascat.h_saf as H_SAF


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H08(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__), 'ascat_test_data',
                                 'hsaf', 'h08')
        self.reader = H_SAF.H08img(data_path)

    def tearDown(self):
        self.reader = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        2010-05-01
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2010, 5, 1), datetime(2010, 5, 1, 12))
        timestamps_should = [datetime(2010, 5, 1, 8, 33, 1)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2010, 5, 1, 8, 33, 1))
        # do not check data content at the moment just shapes and structure
        assert sorted(data.keys()) == sorted(
            ['ssm', 'corr_flag', 'ssm_noise', 'proc_flag'])
        assert lons.shape == (3120, 7680)
        assert lats.shape == (3120, 7680)
        for var in data:
            assert data[var].shape == (3120, 7680)

    def test_image_reading_bbox_empty(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2010, 5, 1, 8, 33, 1), lat_lon_bbox=[45, 48, 15, 18])

        # do not check data content at the moment just shapes and structure
        assert data is None
        assert lons is None
        assert lats is None

    def test_image_reading_bbox(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2010, 5, 1, 8, 33, 1),
            lat_lon_bbox=[60, 70, 15, 25])

        # do not check data content at the moment just shapes and structure
        assert sorted(data.keys()) == sorted(
            ['ssm', 'corr_flag', 'ssm_noise', 'proc_flag'])
        assert lons.shape == (2400, 2400)
        assert lats.shape == (2400, 2400)
        for var in data:
            assert data[var].shape == (2400, 2400)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H07(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(os.path.dirname(__file__),  'ascat_test_data',
                                 'hsaf', 'h07')
        self.reader = H_SAF.H07img(data_path)

    def tearDown(self):
        self.reader = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        2010-05-01
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2010, 5, 1), datetime(2010, 5, 1, 12))
        timestamps_should = [datetime(2010, 5, 1, 8, 33, 1)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2010, 5, 1, 8, 33, 1))

        ssm_should = np.array([51.2, 65.6, 46.2, 56.9, 61.4, 61.5, 58.1, 47.1,
                               72.7, 13.8, 60.9, 52.1, 78.5, 57.8, 56.2, 79.8,
                               67.7, 53.8, 86.5, 29.4, 50.6, 88.8, 56.9, 68.9,
                               52.4, 64.4, 81.5, 50.5, 84., 79.6, 47.4, 79.5,
                               46.9, 60.7, 81.3, 52.9, 84.5, 25.5, 79.2, 93.3,
                               52.6, 93.9, 74.4, 91.4, 76.2, 92.5, 80., 88.3,
                               79.1, 97.2, 56.8])

        lats_should = np.array([70.21162, 69.32506, 69.77325, 68.98149,
                                69.12295, 65.20364, 67.89625, 67.79844,
                                67.69112, 67.57446, 67.44865, 67.23221,
                                66.97207, 66.7103, 66.34695, 65.90996,
                                62.72462, 61.95761, 61.52935, 61.09884,
                                60.54359, 65.60223, 65.33588, 65.03098,
                                64.58972, 61.46131, 60.62553, 59.52057,
                                64.27395, 63.80293, 60.6569, 59.72684,
                                58.74838, 63.42774])

        ssm_mean_should = np.array([0.342,  0.397,  0.402,  0.365,  0.349,
                                    0.354,  0.37,  0.36, 0.445,  0.211,
                                    0.394,  0.354,  0.501,  0.361,  0.366,
                                    0.45, 0.545,  0.329,  0.506,  0.229,
                                    0.404,  0.591,  0.348,  0.433, 0.29,
                                    0.508,  0.643,  0.343,  0.519, 0.61,
                                    0.414,  0.594, 0.399,  0.512,  0.681,
                                    0.457,  0.622,  0.396,  0.572,  0.7,
                                    0.302,  0.722,  0.493,  0.747,  0.521,
                                    0.72,  0.578,  0.718, 0.536,  0.704,
                                    0.466]) * 100

        nptest.assert_allclose(lats[25:-1:30], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               15:-1:20], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][15:-1:20],
                               ssm_mean_should,
                               atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H16(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__),  'ascat_test_data', 'hsaf', 'h16')
        self.reader = H_SAF.H16img(data_path, month_path_str='')

    def tearDown(self):
        self.reader = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        2010-05-01
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2017, 2, 20), datetime(2017, 2, 21))
        timestamps_should = [datetime(2017, 2, 20, 11, 0, 0) +
                             timedelta(minutes=3) * n for n in range(8)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_offset_getting_datetime_boundary(self):
        """
        test getting the image offsets for a known daterange,
        checks if exact datetimes are used
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2017, 2, 20, 11, 3),
            datetime(2017, 2, 20, 11, 12))
        timestamps_should = [datetime(2017, 2, 20, 11, 3, 0) +
                             timedelta(minutes=3) * n for n in range(4)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2017, 2, 20, 11, 15, 0))

        ssm_should = np.array([0., 3.6, 7.8, 8.2, 12.3, 14.7, 21.6, 26.7,
                               30.6, 32.2, 43., 50.5, 46.3, 47.6, 58.])

        lats_should = np.array([-28.25222, -28.21579, -28.1789, -28.14155,
                                -28.10374, -28.06547, -28.02674, -27.98755,
                                -27.94791, -27.90782, -27.86727, -27.82627,
                                -27.78482, -27.74292, -27.70058, ])

        ssm_mean_should = np.array([32.8, 35.1, 37.8, 39.2, 39., 38., 36.2,
                                    39.6, 43.8, 45.6, 46.1, 47., 43.7, 46.,
                                    46.7])

        nptest.assert_allclose(lats[798:813], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               798:813], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][798:813],
                               ssm_mean_should,
                               atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H101(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__),  'ascat_test_data', 'hsaf', 'h101')
        self.reader = H_SAF.H101img(data_path, month_path_str='')

    def tearDown(self):
        self.reader = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        2010-05-01
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2017, 2, 20), datetime(2017, 2, 21))
        timestamps_should = [datetime(2017, 2, 20, 10, 24, 0) +
                             timedelta(minutes=3) * n for n in range(8)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2017, 2, 20, 10, 42, 0))

        ssm_should = np.array([26.1, 31.5, 49.5, 64.3, 80.3, 87.9, 24.5, 18.5,
                               20., 12.9, 6.3, 3.1, 3.5, 5.7, 2., 0., 4.6, 8.,
                               9.6, 11.8])

        lats_should = np.array([56.26292, 56.29267, 56.32112, 56.34827,
                                56.37413, 56.39868, 51.98611, 52.08869,
                                52.19038, 52.29117, 52.39104, 52.48999,
                                52.58802, 52.68512, 52.78127, 52.87648,
                                52.97073, 53.06402, 53.15634, 53.24768])

        ssm_mean_should = np.array([22.8, 27., 31.8, 38.3, 44.3, 52.4, 24.8,
                                    27.8, 24.7, 23.9, 22.8, 25., 25.6, 26.1,
                                    25.9, 26.9, 28.5, 27.8, 24.3, 22.7])

        nptest.assert_allclose(lats[1800:1820], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               1800:1820], ssm_should, atol=0.01)
        nptest.assert_allclose(
            data['Mean Surface Soil Moisture'][1800:1820],
            ssm_mean_should, atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H102(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__),  'ascat_test_data', 'hsaf', 'h102')
        self.reader = H_SAF.H102img(data_path, month_path_str='')

    def tearDown(self):
        self.reader = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        2010-05-01
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2017, 2, 20), datetime(2017, 2, 21))
        timestamps_should = [datetime(
            2017, 2, 20, 10, 24, 0) + timedelta(minutes=3) * n for n in range(8)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2017, 2, 20, 10, 42, 0))

        ssm_should = np.array([45.8, 43.5, 41.7, 42.7, 38.6, 31.1, 23.4, 21.7,
                               23.6, 26.8, 30.5, 30.1, 32.7, 34.7, 35.8, 38.3,
                               46.2, 53.7, 46.2])

        lats_should = np.array([43.16844, 43.21142, 43.25423, 43.29686,
                                43.33931, 43.38158, 43.42367, 43.46558,
                                43.50731, 43.54887, 43.59024, 43.63143,
                                43.67243, 43.71326, 43.7539, 43.79436,
                                43.83463, 43.87472, 43.91463])

        ssm_mean_should = np.array([50.8, 44.5, 36.3, 29.9, 28.7, 29.1, 30.,
                                    30.1, 30.8, 33., 34.9, 35.7, 35.2, 34.2,
                                    32.9, 32.9, 34.5, 32.7, 30.6])

        nptest.assert_allclose(lats[0:19], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               0:19], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][0:19],
                               ssm_mean_should,
                               atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H103(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__),  'ascat_test_data', 'hsaf', 'h103')
        self.reader = H_SAF.H103img(data_path, month_path_str='')

    def tearDown(self):
        self.reader = None

    def test_offset_getting(self):
        """
        test getting the image offsets for a known day
        2010-05-01
        """
        timestamps = self.reader.tstamps_for_daterange(
            datetime(2017, 2, 20), datetime(2017, 2, 21))
        timestamps_should = [datetime(
            2017, 2, 20, 10, 30, 0) + timedelta(minutes=3) * n for n in range(8)]
        assert sorted(timestamps) == sorted(timestamps_should)

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2017, 2, 20, 10, 30, 0))

        ssm_should = np.array([20.4, 15.2, 4.2, 0., 0.7, 6.8, 14.2, 21.9,
                               23.7, 17.4, 14.8, 19.8, 15.4, 17.1, 28.9,
                               28.5, 23.9, 18.1, 25.4])

        lats_should = np.array([8.82896, 8.85636, 8.88373, 8.91107, 8.93837,
                                8.96564, 8.99288, 9.02009, 9.04726, 9.0744,
                                9.1015, 9.12858, 9.15561, 9.18262, 9.20959,
                                9.23653, 9.26343, 9.2903, 9.31713])

        ssm_mean_should = np.array([14., 14., 13.8, 13.5, 13.3, 13.1, 12.8,
                                    12.6, 12.6, 12., 12.1, 12.6, 13.2, 14.1,
                                    13.9, 12.4, 10.9, 10.3, 10.7])

        nptest.assert_allclose(lats[0:19], lats_should, atol=1e-5)
        nptest.assert_allclose(data['Surface Soil Moisture (Ms)'][
                               0:19], ssm_should, atol=0.01)
        nptest.assert_allclose(data['Mean Surface Soil Moisture'][0:19],
                               ssm_mean_should,
                               atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H14(unittest.TestCase):

    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__),  'ascat_test_data', 'hsaf', 'h14')
        self.reader = H_SAF.H14img(data_path, expand_grid=False)
        self.expand_reader = H_SAF.H14img(data_path, expand_grid=True)

    def tearDown(self):
        self.reader = None
        self.expand_reader = None

    def test_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.reader.read(
            datetime(2014, 5, 15))
        assert sorted(data.keys()) == sorted(['SM_layer1_0-7cm',
                                              'SM_layer2_7-28cm',
                                              'SM_layer3_28-100cm',
                                              'SM_layer4_100-289cm'])
        assert lons.shape == (843490,)
        assert lats.shape == (843490,)
        for var in data:
            assert data[var].shape == (843490,)
            assert meta[var]['name']
            assert meta[var]['units'] in ['dimensionless', 'unknown']

    def test_expanded_image_reading(self):
        data, meta, timestamp, lons, lats, time_var = self.expand_reader.read(
            datetime(2014, 5, 15))
        assert sorted(data.keys()) == sorted(['SM_layer1_0-7cm',
                                              'SM_layer2_7-28cm',
                                              'SM_layer3_28-100cm',
                                              'SM_layer4_100-289cm'])
        assert lons.shape == (800, 1600)
        assert lats.shape == (800, 1600)
        for var in data:
            assert data[var].shape == (800, 1600)
            assert meta[var]['name']
            assert meta[var]['units'] in ['dimensionless', 'unknown']


class Test_H25Ts(unittest.TestCase):

    def setUp(self):

        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'h25')
        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')
        static_layer_path = os.path.join(path, 'ascat_test_data', 'hsaf',
                                         'static_layer')

        self.ascat_reader = H_SAF.H25Ts(cdr_path, grid_path,
                                        static_layer_path=static_layer_path)

    def tearDown(self):
        self.ascat_reader.close()

    def test_read(self):

        gpi = 3066159
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi

        np.testing.assert_approx_equal(result.longitude, 19.03533,
                                       significant=4)
        np.testing.assert_approx_equal(result.latitude, 70.05438,
                                       significant=4)

        assert len(result.data) == 7737
        assert result.data.iloc[15].name.to_pydatetime() == \
            datetime(2007, 1, 7, 10, 49, 9, 4)
        assert result.data.iloc[15]['sm'] == 22
        assert result.data.iloc[15]['ssf'] == 1
        assert result.data.iloc[15]['sm_noise'] == 6
        assert result.data.iloc[15]['frozen_prob'] == 0
        assert result.data.iloc[15]['snow_prob'] == 127
        assert result.data.iloc[15]['orbit_dir'].decode('utf-8') == 'D'
        assert result.data.iloc[15]['proc_flag'] == 0

        np.testing.assert_equal(
            result.data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_gldas'], np.nan)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_hwsd'], 0.1078, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_hwsd'], 0.0294, significant=6)

        assert result.topo_complex == 9
        assert result.wetland_frac == 41

        np.testing.assert_approx_equal(
            result.porosity_gldas, np.nan, significant=5)
        np.testing.assert_approx_equal(
            result.porosity_hwsd, 0.49000001, significant=5)


class Test_H108Ts(unittest.TestCase):

    def setUp(self):

        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'h108')
        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')
        static_layer_path = os.path.join(path, 'ascat_test_data', 'hsaf',
                                         'static_layer')

        self.ascat_reader = H_SAF.H108Ts(cdr_path, grid_path,
                                         static_layer_path=static_layer_path)

    def tearDown(self):
        self.ascat_reader.close()

    def test_read(self):

        gpi = 3066159
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi

        np.testing.assert_approx_equal(result.longitude, 19.03533,
                                       significant=4)
        np.testing.assert_approx_equal(result.latitude, 70.05438,
                                       significant=4)

        assert len(result.data) == 8222
        assert result.data.iloc[15].name.to_pydatetime() == \
            datetime(2007, 1, 7, 10, 49, 9, 4)
        assert result.data.iloc[15]['sm'] == 22
        assert result.data.iloc[15]['ssf'] == 2
        assert result.data.iloc[15]['sm_noise'] == 6
        assert result.data.iloc[15]['frozen_prob'] == 0
        assert result.data.iloc[15]['snow_prob'] == 127
        assert result.data.iloc[15]['orbit_dir'].decode('utf-8') == 'D'
        assert result.data.iloc[15]['proc_flag'] == 0

        np.testing.assert_equal(
            result.data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_gldas'], np.nan)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_hwsd'], 0.1078, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_hwsd'], 0.0294, significant=6)

        assert result.topo_complex == 9
        assert result.wetland_frac == 41

        np.testing.assert_equal(
            result.porosity_gldas, np.nan)
        np.testing.assert_approx_equal(
            result.porosity_hwsd, 0.49000001, significant=5)


class Test_H109Ts(unittest.TestCase):

    def setUp(self):

        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'h109')
        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')
        static_layer_path = os.path.join(path, 'ascat_test_data', 'hsaf',
                                         'static_layer')

        self.ascat_reader = H_SAF.H109Ts(cdr_path, grid_path,
                                         static_layer_path=static_layer_path)

    def tearDown(self):
        self.ascat_reader.close()

    def test_read(self):

        gpi = 3066159
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi

        np.testing.assert_approx_equal(result.longitude, 19.03533,
                                       significant=4)
        np.testing.assert_approx_equal(result.latitude, 70.05438,
                                       significant=4)

        assert len(result.data) == 11736
        assert result.data.iloc[15].name.to_pydatetime() == \
            datetime(2007, 1, 7, 10, 49, 9, 379200)
        assert result.data.iloc[15]['sm'] == 27
        assert result.data.iloc[15]['ssf'] == 1
        assert result.data.iloc[15]['sm_noise'] == 5
        assert result.data.iloc[15]['frozen_prob'] == 0
        assert result.data.iloc[15]['snow_prob'] == 127
        assert result.data.iloc[15]['dir'] == 1
        assert result.data.iloc[15]['proc_flag'] == 0
        assert result.data.iloc[15]['corr_flag'] == 16
        assert result.data.iloc[15]['sat_id'] == 3

        np.testing.assert_equal(
            result.data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_gldas'], np.nan)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_hwsd'], 0.1323, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_hwsd'], 0.0245, significant=6)

        assert result.topo_complex == 9
        assert result.wetland_frac == 41

        np.testing.assert_equal(
            result.porosity_gldas, np.nan)
        np.testing.assert_approx_equal(
            result.porosity_hwsd, 0.49000001, significant=5)


class Test_H110Ts(unittest.TestCase):

    def setUp(self):

        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'h110')
        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')
        static_layer_path = os.path.join(path, 'ascat_test_data', 'hsaf',
                                         'static_layer')

        self.ascat_reader = H_SAF.H110Ts(cdr_path, grid_path,
                                         static_layer_path=static_layer_path)

    def tearDown(self):
        self.ascat_reader.close()

    def test_read(self):

        gpi = 3066159
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi

        np.testing.assert_approx_equal(result.longitude, 19.03533,
                                       significant=4)
        np.testing.assert_approx_equal(result.latitude, 70.05438,
                                       significant=4)

        assert len(result.data) == 1148
        assert result.data.iloc[15].name.to_pydatetime() == \
            datetime(2016, 1, 3, 19, 34, 28, 99200)
        assert result.data.iloc[15]['sm'] == 48
        assert result.data.iloc[15]['ssf'] == 1
        assert result.data.iloc[15]['sm_noise'] == 5
        assert result.data.iloc[15]['frozen_prob'] == 0
        assert result.data.iloc[15]['snow_prob'] == 127
        assert result.data.iloc[15]['dir'] == 0
        assert result.data.iloc[15]['proc_flag'] == 0
        assert result.data.iloc[15]['corr_flag'] == 0
        assert result.data.iloc[15]['sat_id'] == 4

        np.testing.assert_equal(
            result.data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_gldas'], np.nan)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_hwsd'], 0.2352, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_hwsd'], 0.0245, significant=6)

        assert result.topo_complex == 9
        assert result.wetland_frac == 41

        np.testing.assert_equal(
            result.porosity_gldas, np.nan)
        np.testing.assert_approx_equal(
            result.porosity_hwsd, 0.49000001, significant=5)


class Test_H111Ts(unittest.TestCase):

    def setUp(self):

        path = os.path.dirname(__file__)

        cdr_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'h111')
        grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')
        static_layer_path = os.path.join(path, 'ascat_test_data', 'hsaf',
                                         'static_layer')

        self.ascat_reader = H_SAF.H111Ts(cdr_path, grid_path,
                                         static_layer_path=static_layer_path)

    def tearDown(self):
        self.ascat_reader.close()

    def test_read(self):

        gpi = 3066159
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi

        np.testing.assert_approx_equal(result.longitude, 19.03533,
                                       significant=4)
        np.testing.assert_approx_equal(result.latitude, 70.05438,
                                       significant=4)

        assert len(result.data) == 13715
        assert result.data.iloc[15].name.to_pydatetime() == \
            datetime(2007, 1, 7, 10, 49, 9, 379200)
        assert result.data.iloc[15]['sm'] == 28
        assert result.data.iloc[15]['ssf'] == 1
        assert result.data.iloc[15]['sm_noise'] == 5
        assert result.data.iloc[15]['frozen_prob'] == 0
        assert result.data.iloc[15]['snow_prob'] == 127
        assert result.data.iloc[15]['dir'] == 1
        assert result.data.iloc[15]['proc_flag'] == 0
        assert result.data.iloc[15]['corr_flag'] == 4
        assert result.data.iloc[15]['sat_id'] == 3

        np.testing.assert_equal(
            result.data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_gldas'], np.nan)

        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_hwsd'], 0.1372, significant=6)
        np.testing.assert_approx_equal(
            result.data.iloc[15]['abs_sm_noise_hwsd'], 0.0245, significant=6)

        assert result.topo_complex == 9
        assert result.wetland_frac == 41

        np.testing.assert_equal(
            result.porosity_gldas, np.nan)
        np.testing.assert_approx_equal(
            result.porosity_hwsd, 0.49000001, significant=5)

    def test_read_2points_cell_switch(self):
        """
        Test reading of two points in two different cells.
        This did not work in the past when the static layer class
        was closed too soon.
        """

        gpi = 3066159
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi
        gpi = 2577735
        result = self.ascat_reader.read(gpi, absolute_sm=True)
        assert result.gpi == gpi


def test_hsaf_ts_class_init():
    """
    Test for initialization of the classes that change nothing but the setup code.
    """

    path = os.path.dirname(__file__)
    grid_path = os.path.join(path, 'ascat_test_data', 'hsaf', 'grid')
    H_SAF.H112Ts('Test', grid_path)
    H_SAF.H113Ts('Test', grid_path)
    H_SAF.H114Ts('Test', grid_path)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from datetime import datetime, timedelta

from pygeobase.io_base import MultiTemporalImageBase


class ASCAT_MultiTemporalImageBase(MultiTemporalImageBase):
    """
    Base class for the lvl1b and lvl2 MultiTemporal data readers
    """

    def __init__(self, *args, **kwargs):
        super(ASCAT_MultiTemporalImageBase, self).__init__(*args, **kwargs)

    def _get_orbit_start_date(self, filename):
        """
        Returns the datetime of the file.

        Parameters
        ----------
        filename : full name (including the path) of the file

        Returns
        -------
        dates : datetime object
            datetime from the filename
        """
        # if your data comes from the EUMETSAT EO Portal this function can
        if self.eo_portal is True:
            filename_base = os.path.basename(filename)
            fln_spl = filename_base.split('_')[4]
            fln_datetime = fln_spl[:-1]
            return datetime.strptime(fln_datetime, self.datetime_format)

        else:
            orbit_start_str = os.path.basename(filename)[
                self.filename_datetime_format[0]:
                self.filename_datetime_format[1]]
            return datetime.strptime(orbit_start_str,
                                     self.filename_datetime_format[2])

    def tstamps_for_daterange(self, startdate, enddate):
        """
        Returns the possible timestamps of the given daterange as a datetime
        array.

        Parameters
        ----------
        startdate : datetime.date or datetime.datetime
            start date
        enddate : datetime.date or datetime.datetime
            end date

        Returns
        -------
        dates : list
            list of datetimes
        """
        file_list = []
        delta_all = enddate - startdate
        timestamps = []

        for i in range(delta_all.days + 1):
            timestamp = startdate + timedelta(days=i)

            files = self._search_files(
                timestamp, custom_templ=self.day_search_str)

            file_list.extend(sorted(files))

        for filename in file_list:
            timestamps.append(self._get_orbit_start_date(filename))

        timestamps = [dt for dt in timestamps if (
            dt >= startdate and dt <= enddate)]
        return timestamps



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import os
import warnings
from gzip import GzipFile
from tempfile import NamedTemporaryFile
from datetime import datetime, timedelta

import numpy as np
try:
    import pygrib
except ImportError:
    warnings.warn(
        'pygrib can not be imported GRIB files (H14) can not be read.')

from pygeobase.io_base import ImageBase
from pygeobase.io_base import MultiTemporalImageBase
from pygeobase.object_base import Image

import ascat.read_native.bufr as bufr_reader
from ascat.read_native.bufr import AscatL2SsmBufr
from ascat.read_native.cdr import AscatNc

if sys.version_info < (3, 0):
    range = xrange


class H08Single(ImageBase):

    def read(self, timestamp=None, lat_lon_bbox=None):
        """
        Read specific image for given datetime timestamp.

        Parameters
        ----------
        filename : string
            filename
        timestamp : datetime.datetime
            exact observation timestamp of the image that should be read
        lat_lon_bbox : list, optional
            list of lat,lon cooridnates of bounding box
            [lat_min, lat_max, lon_min, lon_max]

        Returns
        -------
        data : dict or None
            dictionary of numpy arrays that hold the image data for each
            variable of the dataset, if no data was found None is returned
        metadata : dict
            dictionary of numpy arrays that hold the metadata
        timestamp : datetime.datetime
            exact timestamp of the image
        lon : numpy.array or None
            array of longitudes, if None self.grid will be assumed
        lat : numpy.array or None
            array of latitudes, if None self.grid will be assumed
        time_var : string or None
            variable name of observation times in the data dict, if None all
            observations have the same timestamp
        """
        zipped = False
        if os.path.splitext(self.filename)[1] == '.gz':
            zipped = True

        # for zipped files use an unzipped temporary copy
        if zipped:
            with NamedTemporaryFile(delete=False) as tmp_fid:
                with GzipFile(self.filename) as gz_fid:
                    tmp_fid.write(gz_fid.read())
                filename = tmp_fid.name
        else:
            filename = self.filename

        with bufr_reader.BUFRReader(filename) as bufr:

            lons = []
            ssm = []
            ssm_noise = []
            ssm_corr_flag = []
            ssm_proc_flag = []
            data_in_bbox = True

            for i, message in enumerate(bufr.messages()):
                if i == 0:
                    # first message is just lat, lon extent
                    # check if any data in bbox
                    if lat_lon_bbox is not None:
                        lon_min, lon_max = message[0, 2], message[0, 3]
                        lat_min, lat_max = message[0, 4], message[0, 5]
                        if (lat_lon_bbox[0] > lat_max or
                            lat_lon_bbox[1] < lat_min or
                            lat_lon_bbox[2] > lon_max or
                                lat_lon_bbox[3] < lon_min):
                            data_in_bbox = False
                            break

                elif data_in_bbox:
                    # first 5 elements are there only once, after that,
                    # 4 elements are repeated till the end of the array
                    # these 4 are ssm, ssm_noise, ssm_corr_flag and
                    # ssm_proc_flag each message contains the values for
                    # 120 lons between lat_min and lat_max the grid spacing
                    # is 0.00416667 degrees
                    lons.append(message[:, 0])
                    lat_min = message[0, 1]
                    lat_max = message[0, 2]
                    ssm.append(message[:, 4::4])
                    ssm_noise.append(message[:, 5::4])
                    ssm_corr_flag.append(message[:, 6::4])
                    ssm_proc_flag.append(message[:, 7::4])

        if zipped:
            os.remove(filename)

        if data_in_bbox:

            ssm = np.rot90(np.vstack(ssm)).astype(np.float32)
            ssm_noise = np.rot90(np.vstack(ssm_noise)).astype(np.float32)
            ssm_corr_flag = np.rot90(
                np.vstack(ssm_corr_flag)).astype(np.float32)
            ssm_proc_flag = np.rot90(
                np.vstack(ssm_proc_flag)).astype(np.float32)
            lats_dim = np.linspace(lat_max, lat_min, ssm.shape[0])
            lons_dim = np.concatenate(lons)

            data = {'ssm': ssm,
                    'ssm_noise': ssm_noise,
                    'proc_flag': ssm_proc_flag,
                    'corr_flag': ssm_corr_flag
                    }

            # if there are is a gap in the image it is not a 2D array in
            # lon, lat space but has a jump in latitude or longitude
            # detect a jump in lon or lat spacing

            lon_jump_ind = np.where(np.diff(lons_dim) > 0.00418)[0]

            if lon_jump_ind.size > 1:
                print("More than one jump in longitude")

            if lon_jump_ind.size == 1:
                lon_jump_ind = lon_jump_ind[0]
                diff_lon_jump = np.abs(
                    lons_dim[lon_jump_ind] - lons_dim[lon_jump_ind + 1])
                missing_elements = int(np.round(diff_lon_jump / 0.00416666))
                missing_lons = np.linspace(lons_dim[lon_jump_ind],
                                           lons_dim[lon_jump_ind + 1],
                                           missing_elements,
                                           endpoint=False)

                # fill up longitude dimension to full grid
                lons_dim = np.concatenate([lons_dim[:lon_jump_ind],
                                           missing_lons,
                                           lons_dim[lon_jump_ind + 1:]])

                # fill data with NaN values
                empty = np.empty((lats_dim.shape[0], missing_elements))
                empty.fill(1e38)
                for key in data:
                    data[key] = np.concatenate(
                        [data[key][:, :lon_jump_ind],
                         empty, data[key][:, lon_jump_ind + 1:]], axis=1)

            lat_jump_ind = np.where(np.diff(lats_dim) > 0.00418)[0]

            if lat_jump_ind.size > 1:
                print("More than one jump in latitude")

            if lat_jump_ind.size == 1:
                diff_lat_jump = np.abs(
                    lats_dim[lat_jump_ind] - lats_dim[lat_jump_ind + 1])
                missing_elements = np.round(diff_lat_jump / 0.00416666)
                missing_lats = np.linspace(lats_dim[lat_jump_ind],
                                           lats_dim[lat_jump_ind + 1],
                                           missing_elements,
                                           endpoint=False)

                # fill up longitude dimension to full grid
                lats_dim = np.concatenate(
                    [lats_dim[:lat_jump_ind], missing_lats,
                     lats_dim[lat_jump_ind + 1:]])
                # fill data with NaN values
                empty = np.empty((missing_elements, lons_dim.shape[0]))
                empty.fill(1e38)
                for key in data:
                    data[key] = np.concatenate(
                        [data[key][:lat_jump_ind, :], empty,
                         data[key][lat_jump_ind + 1:, :]], axis=0)

            lons, lats = np.meshgrid(lons_dim, lats_dim)
            # only return data in bbox
            if lat_lon_bbox is not None:
                data_ind = np.where((lats >= lat_lon_bbox[0]) &
                                    (lats <= lat_lon_bbox[1]) &
                                    (lons >= lat_lon_bbox[2]) &
                                    (lons <= lat_lon_bbox[3]))
                # indexing returns 1d array
                # get shape of lats_dim and lons_dim to be able to reshape
                # the 1d arrays to the correct 2d shapes
                lats_dim_shape = \
                    np.where((lats_dim >= lat_lon_bbox[0]) &
                             (lats_dim <= lat_lon_bbox[1]))[0].shape[0]

                lons_dim_shape = \
                    np.where((lons_dim >= lat_lon_bbox[2]) &
                             (lons_dim <= lat_lon_bbox[3]))[0].shape[0]

                lons = lons[data_ind].reshape(lats_dim_shape, lons_dim_shape)
                lats = lats[data_ind].reshape(lats_dim_shape, lons_dim_shape)
                for key in data:
                    data[key] = data[key][data_ind].reshape(
                        lats_dim_shape, lons_dim_shape)

        else:
            lons = None
            lats = None
            data = None

        return Image(lons, lats, data, {}, timestamp)

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class H08img(MultiTemporalImageBase):

    """
    Reads H SAF H08 images. The images have to be uncompressed in the
    following folder structure path - month_path_str (default 'h08_%Y%m_buf')

    For example if path is set to /home/user/hsaf08 and month_path_str
    is left to the default 'h08_%Y%m_buf' then the images for March 2012
    have to be in the folder /home/user/hsaf08/h08_201203_buf/

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on
        the H SAF FTP Server then please specify the string that should be
        used in datetime.datetime.strftime
        Default: 'h08_%Y%m_buf'
    day_search_str: string, optional
        to provide an iterator over all images of a day the method
        _get_possible_timestamps looks for all available images on a day on
        the harddisk. This string is used in datetime.datetime.strftime and
        in glob.glob to search for all files on a day.
        Default : 'h08_%Y%m%d_*.buf'
    file_search_str: string, optional
        this string is used in datetime.datetime.strftime and glob.glob to find
        a 3 minute bufr file by the exact date.
        Default: 'h08_{datetime}*.buf'
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_search_str
        Default: %Y%m%d_%H%M%S
    """

    def __init__(self, path, month_path_str='h08_%Y%m_buf',
                 day_search_str='h08_%Y%m%d_*.buf',
                 file_search_str='h08_{datetime}*.buf',
                 datetime_format='%Y%m%d_%H%M%S',
                 filename_datetime_format=(4, 19, '%Y%m%d_%H%M%S')):

        self.path = path
        self.month_path_str = month_path_str
        self.day_search_str = day_search_str
        self.file_search_str = file_search_str
        self.filename_datetime_format = filename_datetime_format

        super(H08img, self).__init__(path, H08Single,
                                     subpath_templ=[month_path_str],
                                     fname_templ=file_search_str,
                                     datetime_format=datetime_format,
                                     exact_templ=False)

    def _get_orbit_start_date(self, filename):

        orbit_start_str = \
            os.path.basename(filename)[self.filename_datetime_format[0]:
                                       self.filename_datetime_format[1]]

        return datetime.strptime(orbit_start_str,
                                 self.filename_datetime_format[2])

    def tstamps_for_daterange(self, startdate, enddate):
        """
        Get the timestamps as datetime array that are possible for the
        given day, if the timestamps are

        For this product it is not fixed but has to be looked up from
        the hard disk since bufr files are not regular spaced and only
        europe is in this product. For a global product a 3 minute
        spacing could be used as a fist approximation

        Parameters
        ----------
        start_date : datetime.date or datetime.datetime
            start date
        end_date : datetime.date or datetime.datetime
            end date

        Returns
        -------
        dates : list
            list of datetimes
        """
        file_list = []
        delta_all = enddate - startdate
        timestamps = []

        for i in range(delta_all.days + 1):
            timestamp = startdate + timedelta(days=i)

            files = self._search_files(
                timestamp, custom_templ=self.day_search_str)

            file_list.extend(sorted(files))

        for filename in file_list:
            timestamps.append(self._get_orbit_start_date(filename))

        timestamps = [dt for dt in timestamps if dt >=
                      startdate and dt <= enddate]
        return timestamps


class H07img(AscatL2SsmBufr):
    pass


class H16img(AscatL2SsmBufr):

    """
    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on
        the H SAF FTP Server then please specify the string that should be
        used in datetime.datetime.strftime
        Default: 'h16_%Y%m_buf'
    """

    def __init__(self, path, month_path_str='h16_%Y%m_buf'):

        day_search_str = 'h16_%Y%m%d_*.buf'
        file_search_str = 'h16_{datetime}*.buf'

        super(H16img, self).__init__(path, month_path_str=month_path_str,
                                     day_search_str=day_search_str,
                                     file_search_str=file_search_str)


class H101img(AscatL2SsmBufr):

    """
    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on
        the H SAF FTP Server then please specify the string that should
        be used in datetime.datetime.strftime
        Default: 'h101_%Y%m_buf'
    """

    def __init__(self, path, month_path_str='h101_%Y%m_buf'):

        day_search_str = 'h101_%Y%m%d_*.buf'
        file_search_str = 'h101_{datetime}*.buf'
        fn_datetime_fmt = (5, 20, '%Y%m%d_%H%M%S')

        super(H101img, self).__init__(path, month_path_str=month_path_str,
                                      day_search_str=day_search_str,
                                      file_search_str=file_search_str,
                                      filename_datetime_format=fn_datetime_fmt)


class H102img(AscatL2SsmBufr):

    """
    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on
        the H SAF FTP Server then please specify the string that should be
        used in datetime.datetime.strftime
        Default: 'h102_%Y%m_buf'
    """

    def __init__(self, path, month_path_str='h102_%Y%m_buf'):

        day_search_str = 'h102_%Y%m%d_*.buf'
        file_search_str = 'h102_{datetime}*.buf'
        fn_datetime_fmt = (5, 20, '%Y%m%d_%H%M%S')

        super(H102img, self).__init__(path, month_path_str=month_path_str,
                                      day_search_str=day_search_str,
                                      file_search_str=file_search_str,
                                      filename_datetime_format=fn_datetime_fmt)


class H103img(AscatL2SsmBufr):

    """
    Class reading H103 Metop-A ASCAT soil moisture in BUFR format.

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on
        the H SAF FTP Server then please specify the string that should be
        used in datetime.datetime.strftime
        Default: 'h103_%Y%m_buf'
    """

    def __init__(self, path, month_path_str='h103_%Y%m_buf'):

        day_search_str = 'h103_%Y%m%d_*.buf'
        file_search_str = 'h103_{datetime}*.buf'
        fn_datetime_fmt = (5, 20, '%Y%m%d_%H%M%S')

        super(H103img, self).__init__(path, month_path_str=month_path_str,
                                      day_search_str=day_search_str,
                                      file_search_str=file_search_str,
                                      filename_datetime_format=fn_datetime_fmt)


class H14Single(ImageBase):

    """
    Class reading H14 soil moisture in GRIB format.

    Parameters
    ----------
    expand_grid : boolean, optional
        if set the images will be expanded to a 2D image during reading
        if false the images will be returned as 1D arrays on the
        reduced gaussian grid
        Default: True
    metadata_fields: list, optional
        fields of the message to put into the metadata dictionary.
    """

    def __init__(self, filename, mode='r', expand_grid=True,
                 metadata_fields=['units', 'name']):

        self.expand_grid = expand_grid
        self.metadata_fields = metadata_fields
        self.pygrib1 = True

        if int(pygrib.__version__[0]) > 1:
            self.pygrib1 = False

        super(H14Single, self).__init__(filename, mode=mode)

    def read(self, timestamp=None):
        """
        Read specific image for given datetime timestamp.

        Parameters
        ----------
        timestamp : datetime.datetime
            exact observation timestamp of the image that should be read

        Returns
        -------
        data : dict
            dictionary of numpy arrays that hold the image data for each
            variable of the dataset
        metadata : dict
            dictionary of numpy arrays that hold the metadata
        timestamp : datetime.datetime
            exact timestamp of the image
        lon : numpy.array or None
            array of longitudes, if None self.grid will be assumed
        lat : numpy.array or None
            array of latitudes, if None self.grid will be assumed
        time_var : string or None
            variable name of observation times in the data dict, if None all
            observations have the same timestamp
        """

        if self.pygrib1:
            param_names = {'40': 'SM_layer1_0-7cm',
                           '41': 'SM_layer2_7-28cm',
                           '42': 'SM_layer3_28-100cm',
                           '43': 'SM_layer4_100-289cm'}

        else:
            param_names = {
                'SWI1 Soil wetness index in layer 1': 'SM_layer1_0-7cm',
                'SWI2 Soil wetness index in layer 2': 'SM_layer2_7-28cm',
                'SWI3 Soil wetness index in layer 3': 'SM_layer3_28-100cm',
                'SWI4 Soil wetness index in layer 4': 'SM_layer4_100-289cm'}
        data = {}
        metadata = {}

        with pygrib.open(self.filename) as grb:
            for i, message in enumerate(grb):
                message.expand_grid(self.expand_grid)
                if i == 1:
                    lats, lons = message.latlons()
                data[param_names[message['parameterName']]] = message.values

                # read and store metadata
                md = {}
                for k in self.metadata_fields:
                    if message.valid_key(k):
                        md[k] = message[k]
                metadata[param_names[message['parameterName']]] = md

        return Image(lons, lats, data, metadata, timestamp)

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class H14img(MultiTemporalImageBase):

    """
    Class for reading H SAF H14 SM DAS 2 products in grib format.

    The images have to be uncompressed in the following folder structure
    path - month_path_str (default 'h14_%Y%m_grib')

    For example if path is set to /home/user/hsaf14 and month_path_str is
    left to the default 'h14_%Y%m_grib' then the images for March 2012 have
    to be in the folder /home/user/hsaf14/h14_201203_grib/

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on
        the H SAF FTP Server then please specify the string that should be
        used in datetime.datetime.strftime
        Default: 'h14_%Y%m_grib'
    file_str: string, optional
        this string is used in datetime.datetime.strftime to get the
        filename of a H14 daily grib file
        Default: 'H14_%Y%m%d00.grib'
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_str
        Default: %Y%m%d
    """

    def __init__(self, path, month_path_str='h14_%Y%m_grib',
                 file_str='H14_{datetime}00.grib', datetime_format='%Y%m%d',
                 expand_grid=True):

        self.path = path
        self.month_path_str = month_path_str
        self.file_search_str = file_str

        super(H14img, self).__init__(path, H14Single,
                                     subpath_templ=[month_path_str],
                                     fname_templ=file_str,
                                     datetime_format=datetime_format,
                                     ioclass_kws={'expand_grid': expand_grid})


class H25Ts(AscatNc):

    """
    Class reading H25 soil moisture climate data record (CDR).
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H25_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H25Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                    static_layer_path, **kwargs)


class H108Ts(AscatNc):

    """
    Class reading H108 soil moisture climate data record (CDR) extension.
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H108_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H108Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)


class H109Ts(AscatNc):

    """
    Class reading H109 soil moisture climate data record (CDR).
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H109_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H109Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)


class H110Ts(AscatNc):

    """
    Class reading H110 soil moisture climate data record (CDR) extension.
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H110_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H110Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)


class H111Ts(AscatNc):

    """
    Class reading H111 soil moisture climate data record (CDR).
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H111_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H111Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)


class H112Ts(AscatNc):

    """
    Class reading H112 soil moisture climate data record (CDR).
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H112_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H112Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)


class H113Ts(AscatNc):

    """
    Class reading H113 soil moisture climate data record (CDR).
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H113_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H113Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)


class H114Ts(AscatNc):

    """
    Class reading H114 soil moisture climate data record (CDR).
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H114_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H114Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)


class H115Ts(AscatNc):

    """
    Class reading H115 soil moisture climate data record (CDR).
    """

    def __init__(self, cdr_path, grid_path,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None, **kwargs):

        fn_format = 'H115_{:04d}'
        grid_filename = os.path.join(grid_path, grid_filename)

        super(H115Ts, self).__init__(cdr_path, fn_format, grid_filename,
                                     static_layer_path, **kwargs)



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Readers for lvl1b and lvl2 data in bufr format.
"""

import os
from datetime import datetime, timedelta
import warnings
from tempfile import NamedTemporaryFile
from gzip import GzipFile

import numpy as np

from pygeobase.io_base import ImageBase
from pygeobase.io_base import MultiTemporalImageBase
from pygeobase.io_base import IntervalReadingMixin
from pygeobase.object_base import Image

try:
    from pybufr_ecmwf import raw_bufr_file
    from pybufr_ecmwf import ecmwfbufr
    from pybufr_ecmwf import ecmwfbufr_parameters
except ImportError:
    warnings.warn(
        'pybufr-ecmwf can not be imported, H08 and H07 images can '
        'not be read.')


class AscatL1BufrFile(ImageBase):
    """
    Reads ASCAT L1b data in BUFR format.

    Parameters
    ----------
    filename : str
        Filename path.
    mode : str, optional
        Opening mode. Default: r
    msg_name_lookup: dict, optional
        Dictionary mapping bufr msg number to parameter name.
        See :ref:`ascatformattable`.

        Default:

             === =====================================================
             Key Value
             === =====================================================
             4: "Satellite Identifier",
             6: "Direction Of Motion Of Moving Observing Platform",
             16: "Orbit Number",
             17: "Cross-Track Cell Number",
             21: "f_Beam Identifier",
             22: "f_Radar Incidence Angle",
             23: "f_Antenna Beam Azimuth",
             24: "f_Backscatter",
             25: "f_Radiometric Resolution (Noise Value)",
             26: "f_ASCAT KP Estimate Quality",
             27: "f_ASCAT Sigma-0 Usability",
             34: "f_ASCAT Land Fraction",
             35: "m_Beam Identifier",
             36: "m_Radar Incidence Angle",
             37: "m_Antenna Beam Azimuth",
             38: "m_Backscatter",
             39: "m_Radiometric Resolution (Noise Value)",
             40: "m_ASCAT KP Estimate Quality",
             41: "m_ASCAT Sigma-0 Usability",
             48: "m_ASCAT Land Fraction",
             49: "a_Beam Identifier",
             50: "a_Radar Incidence Angle",
             51: "a_Antenna Beam Azimuth",
             52: "a_Backscatter",
             53: "a_Radiometric Resolution (Noise Value)",
             54: "a_ASCAT KP Estimate Quality",
             55: "a_ASCAT Sigma-0 Usability",
             62: "a_ASCAT Land Fraction"
             === =====================================================
    """

    def __init__(self, filename, mode='r', msg_name_lookup=None, **kwargs):
        """
        Initialization of i/o object.

        """
        zipped = False
        if os.path.splitext(filename)[1] == '.gz':
            zipped = True

        # for zipped files use an unzipped temporary copy
        if zipped:
            with NamedTemporaryFile(delete=False) as tmp_fid:
                with GzipFile(filename) as gz_fid:
                    tmp_fid.write(gz_fid.read())
                filename = tmp_fid.name

        super(AscatL1BufrFile, self).__init__(filename, mode=mode,
                                              **kwargs)
        if msg_name_lookup is None:
            msg_name_lookup = {
                4: "Satellite Identifier",
                6: "Direction Of Motion Of Moving Observing Platform",
                16: "Orbit Number",
                17: "Cross-Track Cell Number",
                21: "f_Beam Identifier",
                22: "f_Radar Incidence Angle",
                23: "f_Antenna Beam Azimuth",
                24: "f_Backscatter",
                25: "f_Radiometric Resolution (Noise Value)",
                26: "f_ASCAT KP Estimate Quality",
                27: "f_ASCAT Sigma-0 Usability",
                34: "f_ASCAT Land Fraction",
                35: "m_Beam Identifier",
                36: "m_Radar Incidence Angle",
                37: "m_Antenna Beam Azimuth",
                38: "m_Backscatter",
                39: "m_Radiometric Resolution (Noise Value)",
                40: "m_ASCAT KP Estimate Quality",
                41: "m_ASCAT Sigma-0 Usability",
                48: "m_ASCAT Land Fraction",
                49: "a_Beam Identifier",
                50: "a_Radar Incidence Angle",
                51: "a_Antenna Beam Azimuth",
                52: "a_Backscatter",
                53: "a_Radiometric Resolution (Noise Value)",
                54: "a_ASCAT KP Estimate Quality",
                55: "a_ASCAT Sigma-0 Usability",
                62: "a_ASCAT Land Fraction"}
        self.msg_name_lookup = msg_name_lookup

    def read(self, timestamp=None):
        """
        Read specific image for given datetime timestamp.

        Parameters
        ----------
        timestamp : datetime.datetime
            exact observation timestamp of the image that should be read

        Returns
        -------
        data : dict
            dictionary of numpy arrays that hold the image data for each
            variable of the dataset
        metadata : dict
            dictionary of numpy arrays that hold the metadata
        timestamp : datetime.datetime
            exact timestamp of the image
        lon : numpy.array or None
            array of longitudes, if None self.grid will be assumed
        lat : numpy.array or None
            array of latitudes, if None self.grid will be assumed
        time_var : string or None
            variable name of observation times in the data dict, if None all
            observations have the same timestamp
        """
        # lookup table between names and message number in the BUFR file

        data = {}
        dates = []
        # 13: Latitude (High Accuracy)
        latitude = []
        # 14: Longitude (High Accuracy)
        longitude = []

        with BUFRReader(self.filename) as bufr:
            for message in bufr.messages():
                # read fixed fields
                latitude.append(message[:, 12])
                longitude.append(message[:, 13])
                years = message[:, 6].astype(int)
                months = message[:, 7].astype(int)
                days = message[:, 8].astype(int)
                hours = message[:, 9].astype(int)
                minutes = message[:, 10].astype(int)
                seconds = message[:, 11].astype(int)

                dates.append(
                    julday(months, days, years, hours, minutes, seconds))

                # read optional data fields
                for mid in self.msg_name_lookup:
                    name = self.msg_name_lookup[mid]

                    if name not in data:
                        data[name] = []

                    data[name].append(message[:, mid - 1])

        dates = np.concatenate(dates)
        longitude = np.concatenate(longitude)
        latitude = np.concatenate(latitude)
        n_records = latitude.shape[0]

        for mid in self.msg_name_lookup:
            name = self.msg_name_lookup[mid]
            data[name] = np.concatenate(data[name])

        data['jd'] = dates
        if 'Direction Of Motion Of Moving Observing Platform' in data:
            data['as_des_pass'] = (data[
                "Direction Of Motion Of Moving Observing Platform"
            ] < 270).astype(np.uint8)

        if 'Cross-Track Cell Number' in data:
            if data['Cross-Track Cell Number'].max() == 82:
                data['swath_indicator'] = 1 * (
                    data['Cross-Track Cell Number'] > 41)
            elif data['Cross-Track Cell Number'].max() == 42:
                data['swath_indicator'] = 1 * (
                    data['Cross-Track Cell Number'] > 21)
            else:
                raise ValueError("Unsuspected node number.")
            n_lines = n_records / max(data['Cross-Track Cell Number'])
            data['line_num'] = np.arange(n_lines).repeat(
                max(data['Cross-Track Cell Number']))

        # There are strange elements with a value of 32.32 instead of the
        # typical nan_values
        # Since some elements rly have this value we check the other triplet
        # data of that beam to filter the nan_values out
        beams = ['f', 'm', 'a']
        for beam in beams:
            azi = beam + '_Antenna Beam Azimuth'
            sig = beam + '_Backscatter'
            inc = beam + '_Radar Incidence Angle'
            if azi in data:
                mask_azi = data[azi] == 32.32
                mask_sig = data[sig] == 1.7e+38
                mask_inc = data[inc] == 1.7e+38
                mask = np.all([mask_azi, mask_sig, mask_inc], axis=0)
                data[azi][mask] = 1.7e+38

        # 1 ERS 1
        # 2 ERS 2
        # 3 METOP-1 (Metop-B)
        # 4 METOP-2 (Metop-A)
        # 5 METOP-3 (Metop-C)
        sat_id = {3: 1, 4: 2, 5: 3}

        metadata = {}
        metadata['SPACECRAFT_ID'] = np.int8(
            sat_id[data['Satellite Identifier'][0]])
        metadata['ORBIT_START'] = np.uint32(data['Orbit Number'][0])

        return Image(longitude, latitude, data, metadata,
                     timestamp, timekey='jd')

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class AscatL2SsmBufrFile(ImageBase):
    """
    Reads ASCAT SSM swath files in BUFR format. There are the
    following products:

    - H101 SSM ASCAT-A NRT O 12.5 Metop-A ASCAT NRT SSM orbit geometry
    12.5 km sampling
    - H102 SSM ASCAT-A NRT O 25.0 Metop-A ASCAT NRT SSM orbit geometry
    25 km sampling
    - H16  SSM ASCAT-B NRT O 12.5 Metop-B ASCAT NRT SSM orbit geometry
    12.5 km sampling
    - H103 SSM ASCAT-B NRT O 25.0 Metop-B ASCAT NRT SSM orbit geometry
    25 km sampling
    - H104 SSM ASCAT-C NRT O 12.5 Metop-C ASCAT NRT SSM orbit geometry
    12.5 km sampling
    - H105 SSM ASCAT-C NRT O 25.0 Metop-C ASCAT NRT SSM orbit geometry
    25 km sampling
    - EUMETSAT ASCAT Soil Moisture at 12.5 km Swath Grid - Metop in BUFR format
    - EUMETSAT ASCAT Soil Moisture at 25.0 km Swath Grid - Metop in BUFR format

    Parameters
    ----------
    filename : str
        Filename path.
    mode : str, optional
        Opening mode. Default: r
    msg_name_lookup: dict, optional
        Dictionary mapping bufr msg number to parameter name.
        See :ref:`ascatformattable`.

        Default:

             === =====================================================
             Key Value
             === =====================================================
             4: "Satellite Identifier",
             6: "Direction Of Motion Of Moving Observing Platform",
             16: "Orbit Number",
             17: "Cross-Track Cell Number",
             21: "f_Beam Identifier",
             22: "f_Radar Incidence Angle",
             23: "f_Antenna Beam Azimuth",
             24: "f_Backscatter",
             25: "f_Radiometric Resolution (Noise Value)",
             26: "f_ASCAT KP Estimate Quality",
             27: "f_ASCAT Sigma-0 Usability",
             34: "f_ASCAT Land Fraction",
             35: "m_Beam Identifier",
             36: "m_Radar Incidence Angle",
             37: "m_Antenna Beam Azimuth",
             38: "m_Backscatter",
             39: "m_Radiometric Resolution (Noise Value)",
             40: "m_ASCAT KP Estimate Quality",
             41: "m_ASCAT Sigma-0 Usability",
             48: "m_ASCAT Land Fraction",
             49: "a_Beam Identifier",
             50: "a_Radar Incidence Angle",
             51: "a_Antenna Beam Azimuth",
             52: "a_Backscatter",
             53: "a_Radiometric Resolution (Noise Value)",
             54: "a_ASCAT KP Estimate Quality",
             55: "a_ASCAT Sigma-0 Usability",
             62: "a_ASCAT Land Fraction",
             65: "Surface Soil Moisture (Ms)",
             66: "Estimated Error In Surface Soil Moisture",
             67: "Backscatter",
             68: "Estimated Error In Sigma0 At 40 Deg Incidence Angle",
             69: "Slope At 40 Deg Incidence Angle",
             70: "Estimated Error In Slope At 40 Deg Incidence Angle",
             71: "Soil Moisture Sensitivity",
             72: "Dry Backscatter",
             73: "Wet Backscatter",
             74: "Mean Surface Soil Moisture",
             75: "Rain Fall Detection",
             76: "Soil Moisture Correction Flag",
             77: "Soil Moisture Processing Flag",
             78: "Soil Moisture Quality",
             79: "Snow Cover",
             80: "Frozen Land Surface Fraction",
             81: "Inundation And Wetland Fraction",
             82: "Topographic Complexity"
             === =====================================================
    """

    def __init__(self, filename, mode='r', msg_name_lookup=None, **kwargs):
        """
        Initialization of i/o object.

        """
        super(AscatL2SsmBufrFile, self).__init__(filename, mode=mode,
                                                 **kwargs)
        if msg_name_lookup is None:
            msg_name_lookup = {
                4: "Satellite Identifier",
                6: "Direction Of Motion Of Moving Observing Platform",
                16: "Orbit Number",
                17: "Cross-Track Cell Number",
                21: "f_Beam Identifier",
                22: "f_Radar Incidence Angle",
                23: "f_Antenna Beam Azimuth",
                24: "f_Backscatter",
                25: "f_Radiometric Resolution (Noise Value)",
                26: "f_ASCAT KP Estimate Quality",
                27: "f_ASCAT Sigma-0 Usability",
                34: "f_ASCAT Land Fraction",
                35: "m_Beam Identifier",
                36: "m_Radar Incidence Angle",
                37: "m_Antenna Beam Azimuth",
                38: "m_Backscatter",
                39: "m_Radiometric Resolution (Noise Value)",
                40: "m_ASCAT KP Estimate Quality",
                41: "m_ASCAT Sigma-0 Usability",
                48: "m_ASCAT Land Fraction",
                49: "a_Beam Identifier",
                50: "a_Radar Incidence Angle",
                51: "a_Antenna Beam Azimuth",
                52: "a_Backscatter",
                53: "a_Radiometric Resolution (Noise Value)",
                54: "a_ASCAT KP Estimate Quality",
                55: "a_ASCAT Sigma-0 Usability",
                62: "a_ASCAT Land Fraction",
                65: "Surface Soil Moisture (Ms)",
                66: "Estimated Error In Surface Soil Moisture",
                67: "Backscatter",
                68: "Estimated Error In Sigma0 At 40 Deg Incidence Angle",
                69: "Slope At 40 Deg Incidence Angle",
                70: "Estimated Error In Slope At 40 Deg Incidence Angle",
                71: "Soil Moisture Sensitivity",
                72: "Dry Backscatter",
                73: "Wet Backscatter",
                74: "Mean Surface Soil Moisture",
                75: "Rain Fall Detection",
                76: "Soil Moisture Correction Flag",
                77: "Soil Moisture Processing Flag",
                78: "Soil Moisture Quality",
                79: "Snow Cover",
                80: "Frozen Land Surface Fraction",
                81: "Inundation And Wetland Fraction",
                82: "Topographic Complexity"}
        self.msg_name_lookup = msg_name_lookup

    def read(self, timestamp=None, ssm_masked=False):
        """
        Read specific image for given datetime timestamp.

        Parameters
        ----------
        timestamp : datetime.datetime (optional)
            exact observation timestamp of the image that should be read
        ssm_masked : flag (optional)
            set to True to filter data by ssm values


        Returns
        -------
        data : dict
            dictionary of numpy arrays that hold the image data for each
            variable of the dataset
        metadata : dict
            dictionary of numpy arrays that hold the metadata
        timestamp : datetime.datetime
            exact timestamp of the image
        lon : numpy.array or None
            array of longitudes, if None self.grid will be assumed
        lat : numpy.array or None
            array of latitudes, if None self.grid will be assumed
        time_var : string or None
            variable name of observation times in the data dict, if None all
            observations have the same timestamp
        """
        # lookup table between names and message number in the BUFR file

        data = {}
        dates = []
        # 13: Latitude (High Accuracy)
        latitude = []
        # 14: Longitude (High Accuracy)
        longitude = []

        with BUFRReader(self.filename) as bufr:
            for message in bufr.messages():
                # read fixed fields
                latitude.append(message[:, 12])
                longitude.append(message[:, 13])
                years = message[:, 6].astype(int)
                months = message[:, 7].astype(int)
                days = message[:, 8].astype(int)
                hours = message[:, 9].astype(int)
                minutes = message[:, 10].astype(int)
                seconds = message[:, 11].astype(int)

                dates.append(
                    julday(months, days, years, hours, minutes, seconds))

                # read optional data fields
                for mid in self.msg_name_lookup:
                    name = self.msg_name_lookup[mid]

                    if name not in data:
                        data[name] = []

                    data[name].append(message[:, mid - 1])

        dates = np.concatenate(dates)
        longitude = np.concatenate(longitude)
        latitude = np.concatenate(latitude)
        n_records = latitude.shape[0]

        for mid in self.msg_name_lookup:
            name = self.msg_name_lookup[mid]
            data[name] = np.concatenate(data[name])
            if mid == 74:
                # ssm mean is encoded differently
                data[name] = data[name] * 100

        data['jd'] = dates

        if 'Direction Of Motion Of Moving Observing Platform' in data:
            data['as_des_pass'] = (data[
                "Direction Of Motion Of Moving Observing Platform"]
                <= 270).astype(np.uint8)

        if 'Cross-Track Cell Number' in data:
            if data['Cross-Track Cell Number'].max() == 82:
                data['swath_indicator'] = 1 * (
                    data['Cross-Track Cell Number'] > 41)
            elif data['Cross-Track Cell Number'].max() == 42:
                data['swath_indicator'] = 1 * (
                    data['Cross-Track Cell Number'] > 21)
            else:
                raise ValueError("Unsuspected node number.")
            n_lines = n_records / max(data['Cross-Track Cell Number'])
            data['line_num'] = np.arange(n_lines).repeat(
                max(data['Cross-Track Cell Number']))

        # There are strange elements with a value of 32.32 instead of the
        # typical nan_values
        # Since some elements rly have this value we check the other triplet
        # data of that beam to filter the nan_values out
        beams = ['f', 'm', 'a']
        for beam in beams:
            azi = beam + '_Antenna Beam Azimuth'
            sig = beam + '_Backscatter'
            inc = beam + '_Radar Incidence Angle'
            if azi in data:
                mask_azi = data[azi] == 32.32
                mask_sig = data[sig] == 1.7e+38
                mask_inc = data[inc] == 1.7e+38
                mask = np.all([mask_azi, mask_sig, mask_inc], axis=0)
                data[azi][mask] = 1.7e+38

        # if the ssm_masked is True we mask out data with missing ssm value
        if 65 in self.msg_name_lookup and ssm_masked is True:
            # mask all the arrays based on fill_value of soil moisture
            valid_data = np.where(data[self.msg_name_lookup[65]] != 1.7e+38)
            latitude = latitude[valid_data]
            longitude = longitude[valid_data]
            for name in data:
                data[name] = data[name][valid_data]

        # 1 ERS 1
        # 2 ERS 2
        # 3 METOP-1 (Metop-B)
        # 4 METOP-2 (Metop-A)
        # 5 METOP-3 (Metop-C)
        sat_id = {3: 1, 4: 2, 5: 3}

        metadata = {}

        try:
            metadata['SPACECRAFT_ID'] = np.int8(
                sat_id[data['Satellite Identifier'][0]])
        except KeyError:
            metadata['SPACECRAFT_ID'] = 0

        try:
            metadata['ORBIT_START'] = np.uint32(data['Orbit Number'][0])
        except KeyError:
            metadata['ORBIT_START'] = 0

        return Image(longitude, latitude, data, metadata,
                     timestamp, timekey='jd')

    def resample_data(self, image, index, distance, weights, **kwargs):
        """
        Takes an image and resample (interpolate) the image data to
        arbitrary defined locations given by index and distance.

        Parameters
        ----------
        image : object
            pygeobase.object_base.Image object
        index : np.array
            Index into image data defining a look-up table for data elements
            used in the interpolation process for each defined target
            location.
        distance : np.array
            Array representing the distances of the image data to the
            arbitrary defined locations.
        weights : np.array
            Array representing the weights of the image data that should be
            used during resampling.
            The weights of points not to use are set to np.nan
            This array is of shape (x, max_neighbors)

        Returns
        -------
        image : object
            pygeobase.object_base.Image object
        """
        total_weights = np.nansum(weights, axis=1)

        resOrbit = {}
        # resample backscatter
        for name in image.dtype.names:
            if name in ['Soil Moisture Correction Flag',
                        'Soil Moisture Processing Flag']:
                # The flags are resampled by taking the minimum flag This works
                # since any totally valid observation has the flag 0 and
                # overrides the flagged observations. This is true in cases
                # where the data was set to NaN by the flag as well as when the
                # data was set to 0 or 100. The last image element is the one
                # standing for NaN so we fill it with all flags filled to not
                # interfere with the minimum.
                image[name][-1] = 255
                bits = np.unpackbits(image[name].reshape(
                    (-1, 1)).astype(np.uint8), axis=1)
                resampled_bits = np.min(bits[index, :], axis=1)
                resOrbit[name] = np.packbits(resampled_bits)
            else:
                resOrbit[name] = np.nansum(
                    image[name][index] * weights, axis=1) / total_weights

        return resOrbit

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class AscatL2SsmBufr(MultiTemporalImageBase):
    """
    Class for reading HSAF ASCAT SSM images in bufr format.
    The images have the same structure as the ASCAT 3 minute pdu files
    and these 2 readers could be merged in the future
    The images have to be uncompressed in the following folder structure
    path - month_path_str (default 'h07_%Y%m_buf')

    For example if path is set to /home/user/hsaf07 and month_path_str is left
    to the default 'h07_%Y%m_buf' then the images for March 2012 have to be in
    the folder /home/user/hsaf07/h07_201203_buf/

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        if the files are stored in folders by month as is the standard on the
        HSAF FTP Server then please specify the string that should be used in
        datetime.datetime.strftime Default: 'h07_%Y%m_buf'
    day_search_str: string, optional
        to provide an iterator over all images of a day the
        method _get_possible_timestamps looks for all available images on a day
        on the harddisk. This string is used in datetime.datetime.strftime and
        in glob.glob to search for all files on a day.
        Default : 'h07_%Y%m%d_*.buf'
    file_search_str: string, optional
        this string is used in datetime.datetime.strftime and glob.glob to find
        a 3 minute bufr file by the exact date.
        Default: 'h07_{datetime}*.buf'
    datetime_format: string, optional
        datetime format by which {datetime} will be replaced in file_search_str
        Default: %Y%m%d_%H%M%S
    msg_name_lookup: dict, optional
        Dictionary mapping bufr msg number to parameter name.
        See :ref:`ascatformattable`.

        Default:

             === =====================================================
             Key Value
             === =====================================================
             4:  "Satellite Identifier",
             6:  "Direction Of Motion Of Moving Observing Platform",
             16: "Orbit Number",
             17: "Cross-Track Cell Number",
             21: "f_Beam Identifier",
             22: "f_Radar Incidence Angle",
             23: "f_Antenna Beam Azimuth",
             24: "f_Backscatter",
             25: "f_Radiometric Resolution (Noise Value)",
             26: "f_ASCAT KP Estimate Quality",
             27: "f_ASCAT Sigma-0 Usability",
             34: "f_ASCAT Land Fraction",
             35: "m_Beam Identifier",
             36: "m_Radar Incidence Angle",
             37: "m_Antenna Beam Azimuth",
             38: "m_Backscatter",
             39: "m_Radiometric Resolution (Noise Value)",
             40: "m_ASCAT KP Estimate Quality",
             41: "m_ASCAT Sigma-0 Usability",
             48: "m_ASCAT Land Fraction",
             49: "a_Beam Identifier",
             50: "a_Radar Incidence Angle",
             51: "a_Antenna Beam Azimuth",
             52: "a_Backscatter",
             53: "a_Radiometric Resolution (Noise Value)",
             54: "a_ASCAT KP Estimate Quality",
             55: "a_ASCAT Sigma-0 Usability",
             62: "a_ASCAT Land Fraction",
             65: "Surface Soil Moisture (Ms)",
             66: "Estimated Error In Surface Soil Moisture",
             67: "Backscatter",
             68: "Estimated Error In Sigma0 At 40 Deg Incidence Angle",
             69: "Slope At 40 Deg Incidence Angle",
             70: "Estimated Error In Slope At 40 Deg Incidence Angle",
             71: "Soil Moisture Sensitivity",
             72: "Dry Backscatter",
             73: "Wet Backscatter",
             74: "Mean Surface Soil Moisture",
             75: "Rain Fall Detection",
             76: "Soil Moisture Correction Flag",
             77: "Soil Moisture Processing Flag",
             78: "Soil Moisture Quality",
             79: "Snow Cover",
             80: "Frozen Land Surface Fraction",
             81: "Inundation And Wetland Fraction",
             82: "Topographic Complexity"
             === =====================================================
    """

    def __init__(self, path, month_path_str='h07_%Y%m_buf',
                 day_search_str='h07_%Y%m%d_*.buf',
                 file_search_str='h07_{datetime}*.buf',
                 datetime_format='%Y%m%d_%H%M%S',
                 filename_datetime_format=(4, 19, '%Y%m%d_%H%M%S'),
                 msg_name_lookup=None):
        self.path = path
        self.month_path_str = month_path_str
        self.day_search_str = day_search_str
        self.file_search_str = file_search_str
        self.filename_datetime_format = filename_datetime_format
        super(AscatL2SsmBufr, self).__init__(path, AscatL2SsmBufrFile,
                                             subpath_templ=[month_path_str],
                                             fname_templ=file_search_str,
                                             datetime_format=datetime_format,
                                             exact_templ=False,
                                             ioclass_kws={
                                                 'msg_name_lookup':
                                                     msg_name_lookup})

    def _get_orbit_start_date(self, filename):
        orbit_start_str = \
            os.path.basename(filename)[self.filename_datetime_format[0]:
                                       self.filename_datetime_format[1]]
        return datetime.strptime(orbit_start_str,
                                 self.filename_datetime_format[2])

    def tstamps_for_daterange(self, startdate, enddate):
        """
        Get the timestamps as datetime array that are possible for the
        given day.

        For this product it is not fixed but has to be looked up from
        the hard disk since bufr files are not regular spaced and only
        europe is in this product. For a global product a 3 minute
        spacing could be used as a fist approximation

        Parameters
        ----------
        startdate : datetime.date or datetime.datetime
            start date
        enddate : datetime.date or datetime.datetime
            end date

        Returns
        -------
        dates : list
            list of datetimes
        """
        file_list = []
        delta_all = enddate - startdate
        timestamps = []

        for i in range(delta_all.days + 1):
            timestamp = startdate + timedelta(days=i)

            files = self._search_files(
                timestamp, custom_templ=self.day_search_str)

            file_list.extend(sorted(files))

        for filename in file_list:
            timestamps.append(self._get_orbit_start_date(filename))

        timestamps = [dt for dt in timestamps if (
            startdate <= dt <= enddate)]
        return timestamps


class AscatL2SsmBufrChunked(IntervalReadingMixin, AscatL2SsmBufr):
    """
    Reads BUFR files but does not return them on a file by file basis but in
    bigger chunks. For example it allows to read multiple 3 minute PDU's in
    half orbit chunks of 50 minutes. This speeds up operations like e.g.
    resampling of the data.

    Parameters
    ----------
    chunk_minutes: int, optional
        How many minutes should a chunk of data cover.
    """

    def __init__(self, path, month_path_str='h07_%Y%m_buf',
                 day_search_str='h07_%Y%m%d_*.buf',
                 file_search_str='h07_{datetime}*.buf',
                 datetime_format='%Y%m%d_%H%M%S',
                 filename_datetime_format=(4, 19, '%Y%m%d_%H%M%S'),
                 msg_name_lookup=None, chunk_minutes=50):
        super(AscatL2SsmBufrChunked, self).__init__(
            path,
            month_path_str=month_path_str,
            day_search_str=day_search_str,
            file_search_str=file_search_str,
            datetime_format=datetime_format,
            filename_datetime_format=filename_datetime_format,
            msg_name_lookup=msg_name_lookup,
            chunk_minutes=chunk_minutes)


class BUFRReader(object):
    """
    BUFR reader based on the pybufr-ecmwf package but faster

    Parameters
    ----------
    filename : string
        filename of the bufr file
    kelem_guess : int, optional
        if the elements per variable in as message are known
        please specify here.
        Otherwise the elements will be found out via trial and error
        This works most of the time but is not 100 percent failsafe
        Default: 500
    max_tries : int, optional
        the Reader will try max_tries times to unpack a bufr message.
        Some messages can not be read even if the array sizes are ok.
        Most of the time these files are corrupt.

    """

    def __init__(self, filename, kelem_guess=500, max_tries=10):
        self.bufr = raw_bufr_file.RawBUFRFile()
        self.bufr.open(filename, 'rb')
        self.nr_messages = self.bufr.get_num_bufr_msgs()
        self.max_tries = max_tries

        if 'BUFR_TABLES' not in os.environ:
            path = os.path.split(ecmwfbufr.__file__)[0]
            os.environ["BUFR_TABLES"] = os.path.join(
                path, 'ecmwf_bufrtables' + os.sep)
            # os.environ['PRINT_TABLE_NAMES'] = "false"

        self.size_ksup = ecmwfbufr_parameters.JSUP
        self.size_ksec0 = ecmwfbufr_parameters.JSEC0
        self.size_ksec1 = ecmwfbufr_parameters.JSEC1
        self.size_ksec2 = ecmwfbufr_parameters.JSEC2
        self.size_key = ecmwfbufr_parameters.JKEY
        self.size_ksec3 = ecmwfbufr_parameters.JSEC3
        self.size_ksec4 = ecmwfbufr_parameters.JSEC4

        self.kelem_guess = kelem_guess

    def messages(self):
        """
        Raises
        ------
        IOError:
            if a message cannot be unpacked after max_tries tries

        Returns
        -------
        data : yield results of messages
        """
        count = 0
        for i in np.arange(self.nr_messages) + 1:
            tries = 0

            ksup = np.zeros(self.size_ksup, dtype=np.int)
            ksec0 = np.zeros(self.size_ksec0, dtype=np.int)
            ksec1 = np.zeros(self.size_ksec1, dtype=np.int)
            ksec2 = np.zeros(self.size_ksec2, dtype=np.int)
            ksec3 = np.zeros(self.size_ksec3, dtype=np.int)
            ksec4 = np.zeros(self.size_ksec4, dtype=np.int)

            kerr = 0
            data = self.bufr.get_raw_bufr_msg(i)

            ecmwfbufr.bus012(data[0],  # input
                             ksup,  # output
                             ksec0,  # output
                             ksec1,  # output
                             ksec2,  # output
                             kerr)  # output

            kelem = self.kelem_guess
            ksup_first = ksup[5]
            kvals = ksup_first * kelem
            max_kelem = 500000
            self.init_values = np.zeros(kvals, dtype=np.float64)
            self.cvals = np.zeros((kvals, 80), dtype=np.character)
            # try to expand bufr message with the first guess for
            # kelem
            increment_arraysize = True
            while increment_arraysize:
                cnames = np.zeros((kelem, 64), dtype='|S1')
                cunits = np.zeros((kelem, 24), dtype='|S1')

                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", category=DeprecationWarning)
                    ecmwfbufr.bufrex(data[0],  # input
                                     ksup,  # output
                                     ksec0,  # output
                                     ksec1,  # output
                                     ksec2,  # output
                                     ksec3,  # output
                                     ksec4,  # output
                                     cnames,  # output
                                     cunits,  # output
                                     self.init_values,  # output
                                     self.cvals,  # output
                                     kerr)  # output
                # no error - stop loop
                if kerr == 0 and ksec4[0] != 0:
                    increment_arraysize = False
                # error increase array size and try to unpack again
                else:
                    tries += 1
                    if tries >= self.max_tries:
                        raise IOError('This file seems corrupt')
                    kelem = kelem * 5
                    kvals = ksup_first * kelem

                    if kelem > max_kelem:
                        kelem = kvals / 2
                        max_kelem = kvals

                    self.init_values = np.zeros(kvals, dtype=np.float64)
                    self.cvals = np.zeros((kvals, 80), dtype=np.character)

            decoded_values = ksup[4]
            # set kelem_guess to decoded values of last message
            # only increases reading speed if all messages are the same
            # not sure if this is the best option
            self.kelem_guess = decoded_values
            decoded_msg = ksup[5]
            # calculate first dimension of 2D array
            factor = int(kvals / kelem)

            # reshape and trim the array to the actual size of the data
            values = self.init_values.reshape((factor, kelem))
            values = values[:decoded_msg, :decoded_values]
            count += values.shape[0]

            yield values

    def __enter__(self):
        return self

    def __exit__(self, exc, val, trace):
        self.bufr.close()


def julday(month, day, year, hour=0, minute=0, second=0):
    """
    Julian date from month, day, and year (can be scalars or arrays)
    (function from pytesmo)
    Parameters
    ----------
    month : numpy.ndarray or int32
        Month.
    day : numpy.ndarray or int32
        Day.
    year : numpy.ndarray or int32
        Year.
    hour : numpy.ndarray or int32, optional
        Hour.
    minute : numpy.ndarray or int32, optional
        Minute.
    second : numpy.ndarray or int32, optional
        Second.
    Returns
    -------
    jul : numpy.ndarray or double
        Julian day.
    """
    month = np.array(month)
    day = np.array(day)
    inJanFeb = month <= 2
    jy = year - inJanFeb
    jm = month + 1 + inJanFeb * 12

    jul = np.int32(np.floor(365.25 * jy) +
                   np.floor(30.6001 * jm) + (day + 1720995.0))
    ja = np.int32(0.01 * jy)
    jul += 2 - ja + np.int32(0.25 * ja)

    jul = jul + hour / 24.0 - 0.5 + minute / 1440.0 + second / 86400.0

    return jul



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Readers for data downloaded from EUMETSAT data centre (UMARF)
"""

from ascat.read_native.bufr import AscatL2SsmBufr
from ascat.read_native.bufr import AscatL2SsmBufrChunked
from ascat.read_native.nc import AscatL2SsmNc


class AscatL2Ssm125(AscatL2SsmBufr):
    """
    ASCAT A Level2 Soil Moisture at 12.5 km Swath Grid BUFR files from EUMETSAT

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    """

    def __init__(self, path, month_path_str=''):
        day_search_str = 'M0?-ASCA-ASCSMR02-NA-5.0-%Y%m%d*.bfr'
        file_search_str = 'M0?-ASCA-ASCSMR02-NA-5.0-{datetime}*.bfr'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (25, 39, '%Y%m%d%H%M%S')
        super(AscatL2Ssm125, self).__init__(path, month_path_str=month_path_str,
                                            day_search_str=day_search_str,
                                            file_search_str=file_search_str,
                                            datetime_format=datetime_format,
                                            filename_datetime_format=filename_datetime_format)


class AscatL2Ssm125PDU(AscatL2SsmBufr):
    """
    ASCAT A Level2 Soil Moisture at 12.5 km Swath Grid PDU BUFR files from EUMETSAT

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    """

    def __init__(self, path, month_path_str=''):
        day_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_%Y%m%d*_125_ssm_l2.bin'
        file_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_{datetime}*_125_ssm_l2.bin'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (63, 77, '%Y%m%d%H%M%S')
        super(AscatL2Ssm125PDU, self).__init__(path, month_path_str=month_path_str,
                                               day_search_str=day_search_str,
                                               file_search_str=file_search_str,
                                               datetime_format=datetime_format,
                                               filename_datetime_format=filename_datetime_format)


class AscatL2Ssm125PDUChunked(AscatL2SsmBufrChunked):
    """
    ASCAT A Level2 Soil Moisture at 12.5 km Swath Grid PDU BUFR files from EUMETSAT
    in 50 minute chunks.

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    chunk_minutes: int, optional
        How many minutes should a chunk of data cover.
    """

    def __init__(self, path, month_path_str='', chunk_minutes=100):
        day_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_%Y%m%d*_125_ssm_l2.bin'
        file_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_{datetime}*_125_ssm_l2.bin'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (63, 77, '%Y%m%d%H%M%S')
        super(AscatL2Ssm125PDUChunked, self).__init__(path, month_path_str=month_path_str,
                                                      day_search_str=day_search_str,
                                                      file_search_str=file_search_str,
                                                      datetime_format=datetime_format,
                                                      filename_datetime_format=filename_datetime_format,
                                                      chunk_minutes=chunk_minutes)


class AscatL2Ssm250(AscatL2SsmBufr):
    """
    ASCAT A Level2 Soil Moisture at 25.0 km Swath Grid BUFR files from EUMETSAT

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    """

    def __init__(self, path, month_path_str=''):
        day_search_str = 'M0?-ASCA-ASCSMO02-NA-5.0-%Y%m%d*.bfr'
        file_search_str = 'M0?-ASCA-ASCSMO02-NA-5.0-{datetime}*.bfr'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (25, 39, '%Y%m%d%H%M%S')
        super(AscatL2Ssm250, self).__init__(path, month_path_str=month_path_str,
                                            day_search_str=day_search_str,
                                            file_search_str=file_search_str,
                                            datetime_format=datetime_format,
                                            filename_datetime_format=filename_datetime_format)


class AscatL2Ssm250PDU(AscatL2SsmBufr):
    """
    ASCAT A Level2 Soil Moisture at 25 km Swath Grid PDU BUFR files from EUMETSAT

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    """

    def __init__(self, path, month_path_str=''):
        day_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_%Y%m%d*_250_ssm_l2.bin'
        file_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_{datetime}*_250_ssm_l2.bin'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (63, 77, '%Y%m%d%H%M%S')
        super(AscatL2Ssm250PDU, self).__init__(path, month_path_str=month_path_str,
                                               day_search_str=day_search_str,
                                               file_search_str=file_search_str,
                                               datetime_format=datetime_format,
                                               filename_datetime_format=filename_datetime_format)


class AscatL2Ssm250PDUChunked(AscatL2SsmBufrChunked):
    """
    ASCAT A Level2 Soil Moisture at 25 km Swath Grid PDU BUFR files from EUMETSAT

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    chunk_minutes: int, optional
        How many minutes should a chunk of data cover.
    """

    def __init__(self, path, month_path_str='', chunk_minutes=100):
        day_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_%Y%m%d*_250_ssm_l2.bin'
        file_search_str = 'W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOP?+ASCAT_C_EUM?_{datetime}*_250_ssm_l2.bin'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (63, 77, '%Y%m%d%H%M%S')
        super(AscatL2Ssm250PDUChunked, self).__init__(path, month_path_str=month_path_str,
                                                      day_search_str=day_search_str,
                                                      file_search_str=file_search_str,
                                                      datetime_format=datetime_format,
                                                      filename_datetime_format=filename_datetime_format,
                                                      chunk_minutes=chunk_minutes)


class AscatL2Ssm125Nc(AscatL2SsmNc):
    """
    ASCAT A Level2 Soil Moisture at 12.5 km Swath Grid NetCDF files from EUMETSAT

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    """

    def __init__(self, path, month_path_str=''):
        day_search_str = 'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOP?+ASCAT_C_EUM?_%Y%m%d*_125_ssm_l2.nc'
        file_search_str = 'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOP?+ASCAT_C_EUM?_{datetime}*_125_ssm_l2.nc'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (62, 76, '%Y%m%d%H%M%S')
        super(AscatL2Ssm125Nc, self).__init__(path, month_path_str=month_path_str,
                                              day_search_str=day_search_str,
                                              file_search_str=file_search_str,
                                              datetime_format=datetime_format,
                                              filename_datetime_format=filename_datetime_format)


class AscatL2Ssm250Nc(AscatL2SsmNc):
    """
    ASCAT A Level2 Soil Moisture at 25 km Swath Grid NetCDF files from EUMETSAT

    Parameters
    ----------
    path: string
        path where the data is stored
    month_path_str: string, optional
        If the data is stored in subpaths per year or month then specify the string
        that should be used in datetime.datetime.strftime to get the subpath for a file.
        Default: ''
    """

    def __init__(self, path, month_path_str=''):
        day_search_str = 'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOP?+ASCAT_C_EUM?_%Y%m%d*_250_ssm_l2.nc'
        file_search_str = 'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOP?+ASCAT_C_EUM?_{datetime}*_250_ssm_l2.nc'
        datetime_format = '%Y%m%d%H%M%S'
        filename_datetime_format = (62, 76, '%Y%m%d%H%M%S')
        super(AscatL2Ssm250Nc, self).__init__(path, month_path_str=month_path_str,
                                              day_search_str=day_search_str,
                                              file_search_str=file_search_str,
                                              datetime_format=datetime_format,
                                              filename_datetime_format=filename_datetime_format)
